/* fb.c - (C) 1998 Tomas Berndtsson (tomas@nocrew.org)
 *
 *  This piece of software is released under the terms of the GNU General
 *  Public Licence version 2 or, at your choice, any higher version.
 *  The GNU GPL is included with this package in the file COPYING.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <ofbis.h>
#include "fb.h"

static FB *fb=NULL;
static struct sigaction oldsigs[16];

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

static void signal_handler(int sig)
{
  fprintf(stderr,"fb: signal %d caught and exits program!\n", sig);
  fb_exit();
  exit(1);
}

static void setup_signals()
{
  sigset_t mask;
  struct sigaction sigs;

  sigemptyset(&mask);

  sigs.sa_handler = signal_handler;
  sigs.sa_mask = mask;
  sigs.sa_flags = SA_ONESHOT;
  sigs.sa_restorer = NULL;

  sigaction(SIGHUP, &sigs, &oldsigs[0]);
  sigaction(SIGINT, &sigs, &oldsigs[1]);
  sigaction(SIGQUIT, &sigs, &oldsigs[2]);
  sigaction(SIGILL, &sigs, &oldsigs[3]);
  sigaction(SIGTRAP, &sigs, &oldsigs[4]);
  sigaction(SIGBUS, &sigs, &oldsigs[5]);
  sigaction(SIGFPE, &sigs, &oldsigs[6]);
  sigaction(SIGSEGV, &sigs, &oldsigs[7]);
  sigaction(SIGTERM, &sigs, &oldsigs[8]);
}

static void unset_signals()
{
  sigaction(SIGHUP, &oldsigs[0], NULL);
  sigaction(SIGINT, &oldsigs[1], NULL);
  sigaction(SIGQUIT, &oldsigs[2], NULL);
  sigaction(SIGILL, &oldsigs[3], NULL);
  sigaction(SIGTRAP, &oldsigs[4], NULL);
  sigaction(SIGBUS, &oldsigs[5], NULL);
  sigaction(SIGFPE, &oldsigs[6], NULL);
  sigaction(SIGSEGV, &oldsigs[7], NULL);
  sigaction(SIGTERM, &oldsigs[8], NULL);
}

int fb_setup(int bpp_required)
{
  /* If the framebuffer is already open, don't open it again */
  if(fb)
    return 0;

  fb = FBopen(NULL, FB_OPEN_NEW_VC);
  if(fb == NULL)
    return 1;

  setup_signals();

  if(fb->vinf.bits_per_pixel != bpp_required) {
    fb->vinf.bits_per_pixel = bpp_required;
    fb->vinf.xres_virtual = fb->vinf.xres;
    fb->vinf.yres_virtual = fb->vinf.yres;

    FBputvar(fb);
  }

  return 0;
}

void fb_exit()
{
  if(fb != NULL)
    FBclose(fb);

  fb = NULL;

  unset_signals();
}

FB *fb_getFB()
{
  return fb;
}

void fb_setcolor(int col, int red, int green, int blue)
{
  FBCMAP *fbcmap;

  fbcmap = FBgetcmap(fb);

  if(fbcmap == NULL)
    return;

  col = max(0,min(col,(1<<fb->vinf.bits_per_pixel)-1));

  fbcmap->red[col]   = red;
  fbcmap->green[col] = green;
  fbcmap->blue[col]  = blue;

  fbcmap->start = fbcmap->end = col;
  fbcmap->len = 1;

  FBputcmap(fb, fbcmap);
  FBfreecmap(fbcmap);
}

void fb_setpalette(int *red, int *green, int *blue)
{
  int nr_cols,col;
  FBCMAP *fbcmap;

  fbcmap = FBgetcmap(fb);

  if(fbcmap == NULL)
    return;

  nr_cols = 1<<fb->vinf.bits_per_pixel;

  for(col=0 ; col<nr_cols ; col++) {
    fbcmap->red[col]   = red[col];
    fbcmap->green[col] = green[col];
    fbcmap->blue[col]  = blue[col];
  }

  fbcmap->start = 0;
  fbcmap->end = nr_cols-1;
  fbcmap->len = nr_cols;

  FBputcmap(fb, fbcmap);
  FBfreecmap(fbcmap);
}

void fb_go_tty()
{
  FBVTswitchoriginal(fb);
}

void fb_go_fb()
{
  FBVTswitchfb(fb);
  sleep(1); /* This is an ugly hack! */
}

void fb_waitkey()
{
  FBgetchar(fb);
}

void fb_putpixel(int x, int y, int col)
{
  col = max(0,min(col,(1<<fb->vinf.bits_per_pixel)-1));

  FBputpixel(fb, x, y, col);
}

void fb_clear_area(int x, int y, int w, int h)
{
  FBBLTPBLK *fbb;

  fbb = FBgetbltpblk(fb);

  fbb->b_wd = w;
  fbb->b_ht = h;
  fbb->op_tab = 0x00000000;
  fbb->s_nxwd = 0;
  fbb->s_nxln = 0;
  fbb->s_nxpl = 0;
  fbb->d_xmin = x;
  fbb->d_ymin = y;

  FBbitblt(fb, fbb);

  FBfreebltpblk(fbb);
}

void fb_free_picture(fb_picture *pic)
{
  if(pic != NULL) {
    if(pic->pal_red != NULL)
      free(pic->pal_red);
    if(pic->pal_green != NULL)
      free(pic->pal_green);
    if(pic->pal_blue != NULL)
      free(pic->pal_blue);
    if(pic->data != NULL)
      free(pic->data);
    if(pic->chunkdata != NULL)
      free(pic->chunkdata);
    free(pic);
  }
}

fb_picture *fb_alloc_picture(int width, int height, int bitplanes)
{
  fb_picture *pic;

  if((pic = (fb_picture *)malloc(sizeof(fb_picture))) == NULL) {
    return NULL;
  }
  pic->width = width;
  pic->words = (width+15) >> 4;
  pic->height = height;
  pic->bitplanes = bitplanes;
  pic->size = pic->words*pic->bitplanes*2*pic->height;
  if((pic->data = (unsigned short *)calloc(pic->size, 1)) == NULL) {
    fb_free_picture(pic);
    return NULL;
  }
  if((pic->chunkdata = (unsigned char *)calloc(pic->width*pic->height, 1)) == NULL) {
    fb_free_picture(pic);
    return NULL;
  }
  if((pic->pal_red = (int *)malloc(sizeof(int)*(1<<pic->bitplanes))) == NULL) {
    fb_free_picture(pic);
    return NULL;
  }
  if((pic->pal_green = (int *)malloc(sizeof(int)*(1<<pic->bitplanes))) == NULL) {
    fb_free_picture(pic);
    return NULL;
  }
  if((pic->pal_blue = (int *)malloc(sizeof(int)*(1<<pic->bitplanes))) == NULL) {
    fb_free_picture(pic);
    return NULL;
  }

  return pic;
}

static void putpic(fb_picture *pic, int x, int y)
{
  FBBLTPBLK *fbb;

  fbb = FBgetbltpblk(fb);

  fbb->b_wd = min(pic->width, fb->vinf.xres);
  fbb->b_ht = min(pic->height, fb->vinf.yres);
  fbb->plane_ct = pic->bitplanes;
  fbb->s_xmin = x;
  fbb->s_ymin = y;
  fbb->s_form = pic->data;
  fbb->s_nxwd = pic->bitplanes*2;
  if(fb->finf.type == FB_TYPE_PACKED_PIXELS)
    fbb->s_nxln = pic->width;
  else
    fbb->s_nxln = pic->words*pic->bitplanes*2;
  fbb->s_nxpl = 2;

  FBbitblt(fb, fbb);

  FBfreebltpblk(fbb);
}

static void chunk_to_native(fb_picture *pic)
{
  FBBLOCK chunk, native;

  chunk.width = pic->width;
  chunk.height = pic->height;
  chunk.addr = (void *)pic->chunkdata;

  native.addr = (void *)pic->data;

  FBchunk_to_native(fb, &chunk, &native);
}

int fb_viewpic(fb_picture *pic, int mode, j_settings *js)
{
  int x=0,y=0, new_x=0,new_y=0, ch, quit=0,quit_all=0;

  if(fb_setup(pic->bitplanes)) {
    fprintf(stderr,"Couldn't open framebuffer\n");
    return -1;
  }

  chunk_to_native(pic);

  fb_setpalette(pic->pal_red, pic->pal_green, pic->pal_blue);

  putpic(pic, x, y);

  while(!quit) {
    ch=FBgetchar(fb);
    switch(ch) {
    case 'p':
      fb_setpalette(pic->pal_red, pic->pal_green, pic->pal_blue);
      break;
    case 'u':
      if(pic->height > fb->vinf.yres)
	new_y = y - min(y,fb->vinf.yres);
      break;
    case 'd':
      if(pic->height > fb->vinf.yres)
	new_y = y + min(pic->height-y-fb->vinf.yres,fb->vinf.yres);
      break;
    case 'l':
      if(pic->width > fb->vinf.xres)
	new_x = x - min(x,fb->vinf.xres);
      break;
    case 'r':
      if(pic->width > fb->vinf.xres)
	new_x = x + min(pic->width-x-fb->vinf.xres,fb->vinf.xres);
      break;
    case '\n':
    case '\r':
    case ' ':
      quit=1;
      break;
    case 'q':
      quit=1;
      quit_all=1;
      break;
    default:
    }

    if(x != new_x || y != new_y) {
      x = new_x;
      y = new_y;

      putpic(pic, x, y);
    }
  }

  fb_free_picture(pic);

  if(js->switchfb)
    fb_exit();

  return quit_all;
}

