/* jpeg.c - (C) 1998 Tomas Berndtsson (tomas@nocrew.org)
 *
 *  This piece of software is released under the terms of the GNU General
 *  Public Licence version 2 or, at your choice, any higher version.
 *  The GNU GPL is included with this package in the file COPYING.
 */

#include <stdio.h>
#include <stdlib.h>
#include <jpeglib.h>
#include <setjmp.h>

#include "jpeg.h"
#include "fb.h"

typedef
struct my_error_mgr {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
}
*my_error_ptr;

#define MAXCOLOUR(a) ((int)((a)>65535?65535:(a)))

static inline void j_print_percent(int a, int full)
{
  static int percent=0;
  int new;

  new = a*100 / full;
  if(new != percent) {
    percent = new;
    fprintf(stderr,"\r%d%%", percent);
  }
}

void j_error_exit(j_common_ptr cinfo)
{
  my_error_ptr myerr;

  myerr = (my_error_ptr) cinfo->err;
  longjmp(myerr->setjmp_buffer, 1);
}

fb_picture *j_decompress(FILE *jpegfile, j_settings *js)
{
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr merr;
  int col, line_in, row_stride, x;
  JSAMPARRAY buffer;
  fb_picture *pic;

  cinfo.err = jpeg_std_error(&merr.pub);
  merr.pub.error_exit = j_error_exit;
  if(setjmp(merr.setjmp_buffer)) {
    jpeg_destroy_decompress(&cinfo);
    return NULL;
  }
  
  jpeg_create_decompress(&cinfo);

  jpeg_stdio_src(&cinfo, jpegfile);

  jpeg_read_header(&cinfo, TRUE);

  cinfo.scale_num = js->scale_num;
  cinfo.scale_denom = js->scale_denom;
  cinfo.do_block_smoothing = FALSE;
  switch(js->idct_method) {
  case 0:
    cinfo.dct_method = JDCT_ISLOW;
    break;
  case 1:
    cinfo.dct_method = JDCT_IFAST;
    break;
  case 2:
    cinfo.dct_method = JDCT_FLOAT;
    break;
  default:
    cinfo.dct_method = JDCT_ISLOW;
    break;
  }
  cinfo.quantize_colors = TRUE;
  cinfo.desired_number_of_colors = 1<<js->bitplanes;
  cinfo.two_pass_quantize = js->two_pass ? TRUE : FALSE;
  cinfo.dither_mode = JDITHER_FS;

  jpeg_calc_output_dimensions(&cinfo);

  if((pic = fb_alloc_picture(cinfo.output_width, cinfo.output_height,
			     js->bitplanes)) == NULL) {
    jpeg_destroy_decompress(&cinfo);
    return NULL;
  }

  if(js->verbose) {
    fprintf(stderr,"Width: %d  Height: %d\n", pic->width, pic->height);
    fprintf(stderr,"Quantizing to %d colors with %s-pass quantize.\n", 
	    1<<pic->bitplanes, js->two_pass ? "two" : "one");
  }

  jpeg_start_decompress(&cinfo);

  row_stride = cinfo.output_width * cinfo.output_components;
  buffer = (*cinfo.mem->alloc_sarray)
                ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  line_in=0;
  while (line_in < cinfo.output_height) {
    jpeg_read_scanlines(&cinfo, buffer, 1);

    for(x = 0 ; x < pic->width ; x++)
      pic->chunkdata[pic->width*line_in + x] = buffer[0][x];

    line_in++;

    if(js->verbose) {
      j_print_percent(line_in, cinfo.output_height);
    }
  }
  if(js->verbose)
    fprintf(stderr,"\n");

  for(col=0 ; col < cinfo.actual_number_of_colors ; col++) {
    if(cinfo.out_color_components < 3) {
      pic->pal_red[col]   = 
	pic->pal_green[col] = 
	pic->pal_blue[col]  = MAXCOLOUR((cinfo.colormap[0][col] << 8)*js->brightness);
    } else {
      pic->pal_red[col]   = MAXCOLOUR((cinfo.colormap[0][col] << 8)*js->brightness);
      pic->pal_green[col] = MAXCOLOUR((cinfo.colormap[1][col] << 8)*js->brightness);
      pic->pal_blue[col]  = MAXCOLOUR((cinfo.colormap[2][col] << 8)*js->brightness);
    }
  }
  for( ; col < (1<<pic->bitplanes) ; col++) {
    pic->pal_red[col]   = 0;
    pic->pal_green[col] = 0;
    pic->pal_blue[col]  = 0;
  }

  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  return pic;
}




