/*
 * filein.c: Read, parse and store the contents of a file.
 *
 * This file is part of F-CULT.
 *
 * Copyright (C) 1999 Lauri Hstholm
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * To contact Lauri Hstholm, send e-mail to laurijh@iname.com
 */

#include "fileio.h"

/* Check if the given entry is a dir or a file. Read dir contents
 * if it's a dir, read the file otherwise. */

int read_input(int quieterrors, short listnum, char *infilename)
{
	char origdir[PATH_MAX+1];
	struct stat filestats;

	if (strcmp(infilename, "-")) {
		if (stat(infilename, &filestats)) {
			perror(infilename);
			return 1;
		}
		if (S_ISDIR(filestats.st_mode)) {
			getcwd(origdir, PATH_MAX);
			if (chdir(infilename)) {
				perror(infilename);
				return 1;
			}
			infilename = xrealloc(infilename, NAME_MAX+1);
			getcwd(infilename, PATH_MAX);
			chdir(origdir);
			if (!quiet)
				fprintf(stderr, "Reading directory '%s'..\n", infilename);
			if (read_dir(infilename, listnum))
				return 1;
			inlt = DEFAULTTYPE;
			return 0;
		}
		if (check_access(quieterrors, 0, R_OK, infilename))
			return 1;
	}
	if (!quiet)
		fprintf(stderr, "Reading file '%s'...\n", infilename);
	if (read_file(infilename, listnum))
		return 1;
	return 0;
}

/* If the input file is compressed, we need a pipe to a decompression
 * program to decompress it. Otherwise simply open the file. Then call
 * the type recognizing and list parsing functions. */

int read_file(char *filename, short listnum)
{
	int decomp = 0; /* 1=gzip(/compress), 2=bzip2, 3=(pk)zip */
	char decompstr[NAME_MAX+12] = "";
	FILE *infile;

	if (has_suffix(1, filename, ".gz") || has_suffix(1, filename, ".Z"))
		decomp = GZIP;
	else if (has_suffix(1, filename, ".bz2"))
		decomp = BZIP2;
	else if (has_suffix(1, filename, ".zip"))
		decomp = ZIP;

	switch (decomp) {
		case GZIP:
			strcpy(decompstr, "gzip -cd ");
		case BZIP2:
			if (!strlen(decompstr))
				strcpy(decompstr, "bzip2 -cd ");
		case ZIP:
			if (!strlen(decompstr))
#ifdef __MSDOS__
				strcpy(decompstr, "pkunzip -c ");
#else
				strcpy(decompstr, "unzip -p ");
#endif
			strcat(decompstr, filename);
			if ((infile = popen(decompstr, "r")) == NULL) {
				perror("popen()");
				return 1;
			}
			break;

		default:
			if (!strcmp(filename, "-"))
				infile = stdin;
			else {
				infile = fopen(filename, "r");
				if (infile == NULL) {
					perror("fopen()");
					return 1;
				}
			}
	}

	inlt = gettype(infile);
	if (inlt < 0) {
		fprintf(stderr, "Can't determine list type for '%s'!\n", filename);
		return 1;
	}
	if (parselist(infile, listnum))
		return 1;

	if (decomp)
		pclose(infile);
	else
		fclose(infile);
	return 0;
}

/* Read contents of dirname, push dir entries, store file
 * entries. Returns 0 for everything ok, 1 for error. */

int read_dir(char *dirname, short listnum)
{
	int dirnum;
	int dirstored;
	static int bdnlength;
	char *dirtostore;
	char filepath[PATH_MAX+NAME_MAX+1];
	struct dirent *cf;
	struct stat filestats;
	DIR *dir;

readnext:
	dirnum = dirstored = 0;
	if (!bdnlength)
		bdnlength = strlen(dirname);
	if (verbose)
		fprintf(stderr, "%s\n", dirname);
	dir = opendir(dirname);
	if (dir == NULL) {
		if (!quiet)
			perror(dirname);
		if (top == NULL)
			return 0;
		stack_pop(dirname);
		goto readnext;
	}
	dirtostore = dirname;
	if (stripdirs)
		dirtostore += bdnlength;
	if (dirname[strlen(dirname)-1] != '/')
		strcat(dirname, "/");

	for (;;) {
		do {
			cf = readdir(dir);
			if (cf == NULL) {
				if (!noemptydirs && !dirstored && dirtostore != NULL) {
					if (store_dir(dirtostore) == -1)
						return 1;
				}
				closedir(dir);
				if (top == NULL)
					return 0;
				stack_pop(dirname);
				goto readnext;
			}
		} while (!strcmp(cf->d_name, ".") || !strcmp(cf->d_name, ".."));
		strcpy(filepath, dirname);
		strcat(filepath, cf->d_name);
#ifdef __MSDOS__
		if (stat(filepath, &filestats) && !quiet)
#else
		if (lstat(filepath, &filestats) && !quiet)
#endif
			perror(filepath);
		else {
			if (S_ISDIR(filestats.st_mode)) {
				if (!nonrecursive)
					stack_push(filepath);
			} else {
#ifndef __MSDOS__
				if (S_ISLNK(filestats.st_mode)) {
					if (!nolinks) {
						if (stat(filepath, &filestats) && !quiet)
							perror(filepath);
					} else
						continue;
				}
#endif
				if (!S_ISDIR(filestats.st_mode)) {
					if (!dirstored) {
						dirnum = store_dir(dirtostore);
						if (dirnum == -1)
							return 1;
						dirstored = 1;
					}
					if (store_file(cf->d_name, dirnum, listnum, filestats.st_size, filestats.st_mtime))
						return 1;
				}
			}
		}
	}
}

/* Copy dirtopush on the stack, increase the stack by one. */

void stack_push(char *dirtopush)
{
	struct stack_unit *temp;

	temp = xmalloc(sizeof (struct stack_unit));
	temp->data = xmalloc(strlen(dirtopush)+1);
	strcpy(temp->data, dirtopush);
	temp->next = top;
	top = temp;
}

/* Pop stack data into dirname, decrease the stack by one. */

void stack_pop(char *dirname)
{
	struct stack_unit *temp;

	strcpy(dirname, top->data);
	temp = top;
	top = top->next;
	free(temp->data);
	free(temp);
}

/* Find out list type; return -1 if can't recognize it,
 * one of SLD, SLX, DDD otherwise. */

int gettype(FILE *infile)
{
	int i;
	int j;
	char checkstr[9];
	char line[MAXLINELEN];

	do {
		fgets(line, MAXLINELEN, infile);
		for (i=0; i<MAXLINELEN-8; i++) {
			if (line[i] == '|' || line[i] == '{' || line[i] == '[') {
				for (j=0; j<8; j++)
					checkstr[j] = line[i+j];
				checkstr[8] = '\0';
				if (checkstr[0] == '{')
					checkstr[7] = '\0';
				if (!strcmp(checkstr, SLDID))
					return SLD;
				if (!strcmp(checkstr, SLXID))
					return SLX;
				if (!strcmp(checkstr, DDDID))
					return DDD;
			}
		}
	} while (!feof(infile));
	return -1;
}

/* Read from the file and store file and directory entries. */

int parselist(FILE *infile, short listnum)
{
	int dirnum = 0;
	int dir_or_file = 0;
	long filetime = 0;
	long unsigned filesize = 0;
	char entryname[PATH_MAX+1];
	char line[MAXLINELEN];
	char slx_file_entry_fmt[16];
	char slx_parse_char = '\n';

	dirs = files = 0;
	if (inlt == SLX) {
		do {
			fgets(line, MAXLINELEN, infile);
			if (line[0] == '`' && line[1] == 'P') {
				slx_parse_char = line[12];
				break;
			}
		} while (!feof(infile));
		sprintf(slx_file_entry_fmt, ".%%[^%c]%c%%lu%c%%ld%c", slx_parse_char, slx_parse_char, slx_parse_char, slx_parse_char);
	}

	do {
		fgets(line, MAXLINELEN, infile);
		if (line[0] != '.' && line[0] != '+' && line[0] != '[')
			continue;
		switch (inlt) {
			case SLD:
				if (sscanf(line, "+%s\n", entryname) == 1)
					dir_or_file = 1;
				else if (sscanf(line, ".%s\n", entryname) == 1)
					dir_or_file = 2;
				break;

			case SLX:
				if (sscanf(line, "+%s\n", entryname) == 1)
					dir_or_file = 1;
				else if (sscanf(line, slx_file_entry_fmt, entryname, &filesize, &filetime) == 3)
					dir_or_file = 2;
				break;

			case DDD:
				if (sscanf(line, "[Directory %[^]]\n", entryname) == 1)
					dir_or_file = 1;
				else if (sscanf(line, "[%lu] %s\n", &filesize, entryname) == 2)
					dir_or_file = 2;
				break;
		}

		if (dir_or_file == 1) {
			dirnum = store_dir(entryname);
			if (dirnum == -1)
				return 1;
		} else if (dir_or_file == 2) {
			if (store_file(entryname, dirnum, listnum, filesize, filetime))
				return 1;
		}
		dir_or_file = 0;
	} while (!feof(infile));
	return 0;
}

/* Save directory name in the first available *dircode and update alphadir. */

int store_dir(char *dirname)
{
	int dndiff;
	int i;
	int j;

	if (allinroot) {
		if (!cd) {
			cd = 1;
			dircode = xmalloc(1*sizeof (char *));
			alphadir = xcalloc(1, sizeof (short));
			dircode[0] = xmalloc(strlen(dirname)+1);
			strcpy(dircode[0], dirname);
			alphadir[0] = 0;
		}
		return 0;
	}

	if (!cd) {
		dircode = xmalloc(MAXDIRS*sizeof (char *));
		alphadir = xcalloc(MAXDIRS, sizeof (short));
	} else if (cd == MAXDIRS) {
		fprintf(stderr, "Maximum number of directories (%d) reached!\n", MAXDIRS);
		return -1;
	}
	dircode[cd] = xmalloc(strlen(dirname)+1);
	strcpy(dircode[cd], dirname);
	if (dircode[cd][strlen(dircode[cd])-1] == '/' || dircode[cd][strlen(dircode[cd])-1] == '\\')
		dircode[cd][strlen(dircode[cd])-1] = '\0';
	for (i=0; i<cd; i++) {
		dndiff = strcmp(dircode[cd], dircode[alphadir[i]]);
		if (dndiff < 0) {
			for (j=cd; j>i; j--)
				alphadir[j] = alphadir[j-1];
			alphadir[i] = cd++;
			dirs++;
#ifdef HAVE_LIBNCURSES
			entries++;
#endif
			return cd-1;
		} else if (!dndiff) {
			free(dircode[cd]);
			return alphadir[i];
		}
	}
	alphadir[cd] = cd++;
	dirs++;
#ifdef HAVE_LIBNCURSES
	entries++;
#endif
	return cd-1;
}

/* Store filename in the right position.
 * Not especially pretty, but it works,
 * and isn't too slow either. */

int store_file(char *filename, int dirnum, short listnum, long unsigned filesize, long filetime)
{
	int category;
	int fndiff;
	int i;
	int k;

	if (isupper(filename[0]))
		category = filename[0]-'A';
	else if (islower(filename[0])) {
		if (ignorecase)
			category = filename[0]-'A'-('a'-'A');
		else
			category = filename[0]-'a'+('z'-'a')+1;
	} else if (isdigit(filename[0]))
		category = 2*('z'-'a')+2;
	else
		category = 2*('z'-'a')+3;
	if (suffilter != NULL) {
		for (i=0; i<filters; i++) {
			if (has_suffix(ignorecase, filename, suffilter[i]))
				break;
		}
		if (i == filters)
			return 0;
	}
	if (sfile == NULL) { /* first file of all? */
		sfile = xmalloc(MAXDIRS*sizeof (filestruct ***));
		csf = xmalloc(MAXDIRS*sizeof (short *));
		for (i=0; i<MAXDIRS; i++)
			csf[i] = NULL;
		if (mode != OR)
			activefiles = xmalloc(MAXDIRS*sizeof (char *));
	}
	if (csf[dirnum] == NULL) { /* new dir? */
		csf[dirnum] = xcalloc(CATEGORIES, sizeof (short));
		sfile[dirnum] = xmalloc(CATEGORIES*sizeof (filestruct **));
		for (i=0; i<CATEGORIES; i++)
			sfile[dirnum][i] = NULL;
	} else if (csf[dirnum][category] == MAXFILESINCATEGORY) {
		fprintf(stderr, "Excess files (%d) for one category.\n", MAXFILESINCATEGORY);
		return 1;
	} else if (csf[dirnum][category] && csf[dirnum][category]%INITFILESINCATEGORY == 0)
		sfile[dirnum][category] = xrealloc(sfile[dirnum][category], (INITFILESINCATEGORY+csf[dirnum][category])*sizeof (filestruct *));
/* End of initialization. */

	if (sfile[dirnum][category] == NULL)
		sfile[dirnum][category] = xmalloc(INITFILESINCATEGORY*sizeof (filestruct *));
	else {
/* Scan forward in the category until an
 * entry 'bigger' than filename is found: */
		for (i=0; i<csf[dirnum][category]; i++) {
			if (ignorecase)
				fndiff = strcasecmp(filename, sfile[dirnum][category][i]->filename);
			else
				fndiff = strcmp(filename, sfile[dirnum][category][i]->filename);
			if (fndiff < 0) {
/* Shift all files after the current one by 1: */
				for (k=csf[dirnum][category]; k>i; k--)
					sfile[dirnum][category][k] = sfile[dirnum][category][k-1];
/* Store given info in place of the entry before which we're supposed to go: */
storeit:
				sfile[dirnum][category][i] = xmalloc(sizeof (filestruct));
				sfile[dirnum][category][i]->dirnum = dirnum;
				sfile[dirnum][category][i]->listnum = listnum;
				sfile[dirnum][category][i]->filesize = filesize;
				sfile[dirnum][category][i]->filetime = filetime;
				sfile[dirnum][category][i]->filename = xmalloc(strlen(filename)+1);
				strcpy(sfile[dirnum][category][i]->filename, filename);
				csf[dirnum][category]++;
				files++;
				if (!listnum) {
#ifdef HAVE_LIBNCURSES
					entries++;
#endif
					if (mode != OR)
						activefiles[dirnum]++;
				}
				return 0;
			} else if (!fndiff) {
/* 2 files with the same name? This is where the mode of operation matters;
 * in the XOR mode we mark the existing entry as 'non-existent', in the NOT
 * mode we do that only if it has listnum > 0, in the OR mode we replace
 * the existing file with the new one and in the AND mode we make it active
 * by changing its listnum to 0. */
				switch (mode) {
					case XOR:
						sfile[dirnum][category][i]->listnum = -1;
						if (sfile[dirnum][category][i]->listnum)
							activefiles[dirnum]--;
						return 0;

					case NOT:
						if (sfile[dirnum][category][i]->listnum > 0) {
							sfile[dirnum][category][i]->listnum = -1;
							activefiles[dirnum]--;
						}
						return 0;

					case OR:
						free(sfile[dirnum][category][i]->filename);
						free(sfile[dirnum][category][i]);
						csf[dirnum][category]--;
						files--;
						goto storeit;

					case AND:
						sfile[dirnum][category][i]->listnum = 0;
						activefiles[dirnum]++;
						return 0;
				}
			}
		}
	}
/* This is for fall-throughs and for the first file in the dir: */
	i = csf[dirnum][category];
	goto storeit;
}
