/*
 * fileout.c: Write to a file.
 *
 * This file is part of F-CULT.
 *
 * Copyright (C) 1999 Lauri Hstholm
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * To contact Lauri Hstholm, send e-mail to laurijh@iname.com
 */

#include "fileio.h"

/* This is basically read_input() reversed. Open a pipe to a compressor
 * program if compression is requested, otherwise create a file, then
 * call outputlist() to actually write to the file. */

int write_output(char *filename)
{
	int i;
	char *filename_modified;
	char compstr[NAME_MAX+20] = "";
	FILE *outfile;

	for (i=0; i<cd; i++) {
		if (any_active_files(i))
			break;
	}
	if (i == cd) {
		fprintf(stderr, "There is nothing to write to '%s'.\n", filename);
		return 0;
	}
	filename_modified = xmalloc(strlen(filename)+5);
	strcpy(filename_modified, filename);
	if (!quiet)
		fprintf(stderr, "Writing to '%s'..\n", filename_modified);
	switch (compressor) {
		case GZIP:
			strcpy(compstr, "gzip -fq");
			if (!has_suffix(1, filename, ".gz"))
				strcat(filename_modified, ".gz");

		case BZIP2:
			if (!strlen(compstr)) {
				strcpy(compstr, "bzip2");
				if (!has_suffix(1, filename, ".bz2"))
					strcat(filename_modified, ".bz2");
			}

		case ZIP:
			if (!strlen(compstr)) {
				strcpy(compstr, "zip -q");
				if (!has_suffix(1, filename, ".zip"))
					strcat(filename_modified, ".zip");
			}
			if (strcmp(filename, "-")) {
				strcat(compstr, " > ");
				strcat(compstr, filename_modified);
		       	}
			outfile = popen(compstr, "w");
			if (outfile == NULL) {
				perror(filename_modified);
				return 1;
			}
			break;

		default:
			if (!strcmp(filename, "-"))
				outfile = stdout;
			else {
				outfile = fopen(filename, "w");
				if (outfile == NULL) {
					perror(filename);
					return 1;
				}
			}
	}

	appendheader(outfile);
	files = dirs = 0;
	outputlist(outfile);
	appendfooter(outfile);

	if (compressor)
		pclose(outfile);
	else
		fclose(outfile);
	if (!quiet)
		fprintf(stderr, "Wrote %d directory entries, %lu file entries to '%s'\n", dirs, files, filename_modified);
	free(filename_modified);
	return 0;
}

/* For each dir, append the dir name and every file in it. */

void outputlist(FILE *outfile)
{
	int category;
	int d;
	int dir;
	int i;

	for (d=0; d<cd; d++) {
		dir = alphadir[d];
		if (!any_active_files(dir))
			continue;
		append_dir(outfile, dircode[dir]);
		if (csf[dir] == NULL)
			continue;
		for (category=0; category<CATEGORIES; category++) {
			if (sfile[dir][category] == NULL)
				continue;
			for (i=0; i<csf[dir][category]; i++) {
				if (mode == OR || !sfile[dir][category][i]->listnum)
					append_file(outfile, sfile[dir][category][i]->filename, sfile[dir][category][i]->filesize, sfile[dir][category][i]->filetime);
			}
		}
	}
}

/* Output file name and possibly size and time. */

void append_file(FILE *outfile, char *filename, long unsigned filesize, long filetime)
{
	files++;
	switch (outlt) {
		case SLD:
			fprintf(outfile, ".%s\n", filename);
			break;

/* We always use '|' as the slx parsing char: */
		case SLX:
			fprintf(outfile, ".%s|%lu|%ld|\n", filename, filesize, filetime);
			break;

		case DDD:
			fprintf(outfile, "[%09lu] %s\n", filesize, filename);
			break;
	}
}

/* Output dir name.. */

void append_dir(FILE *outfile, char *dirname)
{
	char *dirname_modified;

	dirs++;
	dirname_modified = xmalloc(strlen(dirname)+1);
	strcpy(dirname_modified, dirname);
	dirname_modified = cleandirname(dirname_modified);
	switch (outlt) {
		case SLD:	/* sld and slx have identical dir entries */
		case SLX:
			fprintf(outfile, "+%s\n", dirname_modified);
			break;

		case DDD:
			fprintf(outfile, "\n[Directory %s]\n", dirname_modified);
	}
	free(dirname_modified);
}

/* ..header.. */

void appendheader(FILE *outfile)
{
	int filterlistlength = 0;
	int i;
	char datestr[11];	/* E.g. "04/09/1999" */
	char *dirname_modified;
	char *filterlist = NULL;
	char timestr[26];
	struct tm *ti;
	time_t epoch = time(NULL);

	switch (outlt) {
		case SLD:
			time(&epoch);
			strcpy(timestr, ctime(&epoch));
/* ctime() appends a newline: */
			timestr[strlen(timestr)-1] = '\0';
			if (suffilter != NULL) {
				for (i=0; i<filters; i++) {
					filterlistlength += strlen(suffilter[i])+1; /* +1 for the commas */
				}
				filterlist = xmalloc(filterlistlength+1);
				strcpy(filterlist, suffilter[0]);
				for (i=1; i<filters; i++) {
					strcat(filterlist, ",");
					strcat(filterlist, suffilter[i]);
				}
			}
			fprintf(outfile, "%s|%s|%s|%s|%s|%s|%s|%d|%s|\n\n\n", SLDCOMMENTS, SLDFIELD0, SLDFIELD1, SLDFIELD2, timestr, SLDFIELD4, (suffilter ? "FALSE" : "TRUE"), dirs, (suffilter ? filterlist : ""));
			break;

		case SLX:
			fprintf(outfile, "%s%s%s%s%s%s%s%s%s%s%s%s%ld\n%s%lu\n%s%d\n%s%s%s%s", SLXCOMMENTS, SLXFIELD0, SLXFIELD1, SLXFIELD2, SLXFIELD3, SLXFIELD4, SLXFIELD5, VERSION, SLXFIELD6, SLXFIELD7, SLXFIELD8, SLXFIELD9, (long)epoch, SLXFIELD10, files, SLXFIELD11, dirs, SLXFIELD12, SLXFIELD13, SLXFIELD14, SLXFIELD15);
			break;

		case DDD:
			ti = localtime(&epoch);
/* localtime() gives the year as >=100 if it's on the 21st century: */
			sprintf(datestr, "%02d/%02d/%d", ti->tm_mon+1, ti->tm_mday, 1900+ti->tm_year);
/* It also gives 24-h format: */
			sprintf(timestr, "%02d:%02d%s", (ti->tm_hour > 12) ? ti->tm_hour-12 : ti->tm_hour, ti->tm_min, (ti->tm_hour > 12) ? "pm" : "am");
			fprintf(outfile, "%s%s%s%s%s\n%s%s\n%s%d\n%s%lu\n\n[Set]\n", DDDCOMMENTS, DDDFIELD0, DDDFIELD1, DDDFIELD2, datestr, DDDFIELD3, timestr, DDDFIELD4, dirs, DDDFIELD5, files);

			for (i=0; i<cd; i++) {
				if (any_active_files(alphadir[i])) {
					dirname_modified = xmalloc(strlen(dircode[alphadir[i]])+1);
					strcpy(dirname_modified, dircode[alphadir[i]]);
					dirname_modified = cleandirname(dirname_modified);
					fprintf(outfile, "%s\n", dirname_modified);
					free(dirname_modified);
				}
			}
			break;
	}
}

/* ..and footer. */

void appendfooter(FILE *outfile)
{
	switch (outlt) {
		case SLX:
			fprintf(outfile, "%s", SLXFOOTER);
			break;

		case DDD:
			fprintf(outfile, "%s", DDDFOOTER);
			break;
	}
}

/* Return 1 if there are any 'active'
 * files in dir dirnum, 0 otherwise. */

int any_active_files(int dirnum)
{
	if (mode != OR) {
		if (activefiles != NULL) {
			if (activefiles[dirnum])
				return 1;
		}
		return 0;
	} else if (csf != NULL) {
		if (csf[dirnum] != NULL || !noemptydirs)
			return 1;
	}
	return 0;
}

/* Remove trailing '/' or '\', convert '/'s to
 * '\'s if dosdirs != 0, else vice versa and
 * add dirprefix. May change dirname's address,
 * returns the new one. */

char *cleandirname(char *dirname)
{
	int i;
	char *tempbuf;

	tempbuf = xmalloc(strlen(dirname)+1);
	for (i=0; i<strlen(dirname); i++) {
		if (dosdirs) {
			if (dirname[i] == '/')
				tempbuf[i] = '\\';
			else
				tempbuf[i] = dirname[i];
		} else {
		       	if (dirname[i] == '\\')
				tempbuf[i] = '/';
			else
				tempbuf[i] = dirname[i];
		}
	}
	tempbuf[i] = '\0';
	if (dirprefix != NULL) {
		dirname = xrealloc(dirname, strlen(tempbuf)+strlen(dirprefix)+1);
		strcpy(dirname, dirprefix);
		strcat(dirname, tempbuf);
	} else
		strcpy(dirname, tempbuf);
	free(tempbuf);
	if (dirname[1] && (dirname[strlen(dirname)-1] == '/' || dirname[strlen(dirname)-1] == '\\'))
		dirname[strlen(dirname)-1] = '\0';
	return dirname;
}
