/*
 * main.c: Code common to all working modes and list formats.
 * 
 * This file is part of F-CULT.
 * 
 * Copyright (C) 1999 Lauri Hstholm
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * To contact Lauri Hstholm, send e-mail to laurijh@iname.com
 */

#include "main.h"

/* Call the functions to parse options, check
 * files and to read and write them. */

int main(int argc, char **argv)
{
	if (parse_options(argc, argv))
		exit(1);
	if (mode) {
		if (check_files(argc, argv))
			exit(1);
		if (read_write_files())
			exit(1);
		exit(0);
	}
	exit(1);
}

/* Parse command-line options and set variables
 * accordingly. */

int parse_options(int argc, char **argv)
{
	char c;
	char *exec_name = argv[0];
#ifdef HAVE_LIBNCURSES
	char short_options[] = "Aa:Bbc:deFf:hIilnNOo:pqrSs::t:VvXZz";
#else
	char short_options[] = "Aa:Bc:deFf:hIilmnNOo:qrSs::t:VvXZz";
#endif
#ifdef HAVE_GETOPT_H
	struct option const long_options[] = {
		{"add-prefix", 1, NULL, 'a'},
		{"all-in-root", 0, NULL, 'r'},
		{"allow-trailing", 0, NULL, 't'},
		{"and", 0, NULL, 'A'},
		{"browse", 0, NULL, 'B'},
		{"bzip2", 0, NULL, 'b'},
		{"combine", 0, NULL, 'O'},
		{"compare", 0, NULL, 'X'},
		{"config-file", 1, NULL, 'c'},
		{"dos-dirs", 0, NULL, 'd'},
		{"duplicates", 0, NULL, 'A'},
		{"filter", 1, NULL, 'f'},
		{"force-overwrite", 0, NULL, 'F'},
		{"gzip", 0, NULL, 'z'},
		{"help", 0, NULL, 'h'},
		{"ignore-case", 0, NULL, 'I'},
		{"ignore-symlinks", 0, NULL, 'i'},
		{"license", 0, NULL, 'l'},
		{"missing", 0, NULL, 'N'},
		{"no-empty-dirs", 0, NULL, 'e'},
		{"no-subdirs", 0, NULL, 'n'},
		{"not", 0, NULL, 'N'},
		{"or", 0, NULL, 'O'},
		{"output-to", 1, NULL, 'o'},
		{"plain-gfx", 0, NULL, 'p'},
		{"quiet", 0, NULL, 'q'},
		{"start-from", 2, NULL, 's'},
		{"strip-dirs", 0, NULL, 'S'},
		{"type", 1, NULL, 'T'},
		{"verbose", 0, NULL, 'v'},
		{"version", 0, NULL, 'V'},
		{"xor", 0, NULL, 'X'},
		{"zip", 0, NULL, 'Z'},
		{NULL, 0, NULL, 0}
	}; 
#endif

	startfrom = -1;

#ifdef HAVE_GETOPT_H
	while ((c = getopt_long(argc, argv, short_options, long_options, NULL)) != EOF) {
#else
	while ((c = getopt(argc, argv, short_options)) != EOF) {
#endif
		switch (c) {
			case 0:
				break;

			case 'A':
				if (mode && mode != NOT) {
					errmsg(FCULT_EXTRA_MODE, exec_name);
					return 1;
				}
				mode = AND;
				break;

#ifdef HAVE_LIBNCURSES
			case 'B':
				browse = 1;
				break;

			case 'c':
				if (!configfile)
					configfile = xmalloc(strlen(optarg)+1);
				strcpy(configfile, optarg);

			case 'p':
				plaingfx = 1;
				break;
#endif

			case 'F':
				foverwrite = 1;
				break;

			case 'I':
				ignorecase = 1;
				break;

			case 'N':
				if (mode && mode != NOT) {
					errmsg(FCULT_EXTRA_MODE, exec_name);
					return 1;
				}
				mode = NOT;
				break;

			case 'O':
				if (mode && mode != OR) {
					errmsg(FCULT_EXTRA_MODE, exec_name);
					return 1;
				}
				mode = OR;
				break;

			case 'S':
				stripdirs = 1;
				break;

			case 'T':
				if (!strcmp(optarg, "0") || !strcasecmp(optarg, "sld"))
					outlt = SLD;
				else if (!strcmp(optarg, "1") || !strcasecmp(optarg, "slx"))
					outlt = SLX;
				else if (!strcmp(optarg, "2") || !strcasecmp(optarg, "ddd"))
					outlt = DDD;
				else {
					errmsg(FCULT_INV_TYPE, exec_name);
					return 1;
				}
				break;

			case 'V':
				fprintf(stderr, "F-CULT version %s\n", VERSION);
				return 0;

			case 'X':
				if (mode && mode != XOR) {
					errmsg(FCULT_EXTRA_MODE, exec_name);
					return 1;
				}
				mode = XOR;
				break;

			case 'Z':
				compressor = ZIP;
				break;

			case 'a':
				if (!dirprefix)
					dirprefix = xmalloc(strlen(optarg)+1);
				strcpy(dirprefix, optarg);
				break;

			case 'b':
				compressor = BZIP2;
				break;

			case 'd':
				dosdirs = 1;
				break;

			case 'e':
				noemptydirs = 1;
				break;

			case 'f':
				if (!suffilter)
					parse_filters(optarg);
				break;

			case 'h':
				show_help(exec_name);
				return 0;

			case 'i':
				nolinks = 1;
				break;

			case 'l':
				fprintf(stderr, LICENSESTR);
				return 0;

			case 'n':
				nonrecursive = 1;
				break;

			case 'o':
				if (!outputname)
					outputname = xmalloc(strlen(optarg)+1);
				strcpy(outputname, optarg);
				break;

			case 'q':
				quiet = 1;
				break;

			case 'r':
				allinroot = 1;
				break;

			case 's':
				if (!optarg)
					startfrom = 1;
				else
					startfrom = atoi(optarg);
				break;

			case 't':
				allowtrailing = 1;
				break;

			case 'v':
				verbose = 1;
				break;

			case 'z':
				compressor = GZIP;
				break;

			default:
				errmsg(0, exec_name);
				return 1;
		}
	}

	if (!mode) {
		errmsg(0, exec_name);
		return 1;
	}
	if (outlt < 0)
		outlt = DEFAULTTYPE;
	return 0;
}

/* Sanity check output file, all input files and browser
 * config file if mode=browse, then store the file names. */

int check_files(int argc, char **argv)
{
	int i;
	int j;

	inputfiles = argc-optind;
	if (!outputname) {
		outputname = xmalloc(2);
		strcpy(outputname, "-");
	} else if (check_access(0, foverwrite, W_OK, outputname))
		return 1;
	inputname = xmalloc(inputfiles*sizeof (char *));
	for (i=0; i<inputfiles; i++) {
		inputname[i] = xmalloc(strlen(argv[optind])+1);
		strcpy(inputname[i], argv[optind++]);
		if (quiet && inputfiles > 1) {
			if (check_access(1, 0, R_OK, inputname[i])) {
				free(inputname[i]);
				inputfiles--;
				i--;
			}
		} else {
			if (check_access(0, 0, R_OK, inputname[i])) {
				free(inputname[i]);
				inputfiles--;
				i--;
			}
		}
		for (j=0; j<i; j++) {
			if (!strcmp(inputname[i], inputname[j])) {
				if (!quiet)
					fprintf(stderr, "Ignoring repeated occurrence of '%s'.\n", inputname[i]);
				free(inputname[i]);
				inputfiles--;
				i--;
			}
		}
	}
	if (!inputfiles) {
		inputfiles = 1;
		inputname = xmalloc(1*sizeof (char *));
		inputname[0] = xmalloc(2);
		strcpy(inputname[0], "-");
	}

#ifdef HAVE_LIBNCURSES
	if (browse) {
/* Remove this loop and the respective error message
 * from errmsg() when reading from stdin for the browser
 * is fixed: */
		for (i=0; i<inputfiles; i++) {
			if (!strcmp(inputname[i], "-")) {
				errmsg(FCULT_BROWSE_STDIN, argv[0]);
				return 1;
			}
		}
		if (!configfile) {
			configfile = xmalloc(strlen(getenv("HOME"))+strlen("/.fcultrc")+1);
			strcpy(configfile, getenv("HOME"));
			strcat(configfile, "/.fcultrc");
		} else if (!strcmp(configfile, "-")) {
			errmsg(FCULT_STDIN_CONF, argv[0]);
			return 1;
		}
		if (check_access(0, 1, R_OK|W_OK, configfile))
			return 1;
	}
#endif
	return 0;
}

/* Read files, do the xor-with-complete thingy if mode
 * of operation is NOT, then browse the list if browse=1,
 * otherwise write the output file. */

int read_write_files()
{
	int i;
	int validfiles = inputfiles;

	for (i=0; i<inputfiles; i++) {
		if (read_input((quiet && validfiles > 1) ? 1 : 0, i+(mode == NOT || mode == AND ? 1 : 0), inputname[i]))
			validfiles--;
		if (!outlt)
			outlt = inlt;
	}
	if (!validfiles)
		return 1;
	if (mode == NOT || mode == AND) {
		inputfiles++;
		if (mode == NOT)
			not_main_loop();
	}
#ifdef HAVE_LIBNCURSES
	if (browse) {
		entries += 3;
		fptr = xmalloc(((entries/FPTRBLOCKSIZE)+1)*sizeof (struct filepointer *));
		for (i=0; i<(entries/FPTRBLOCKSIZE)+1; i++)
			fptr[i] = xmalloc(FPTRBLOCKSIZE*sizeof (struct filepointer));
		browse_indexize_files();
		free(alphadir);
		if (entries <= 3) {
			fprintf(stderr, "Empty file, nothing to browse.\n");
			return 0;
		}
		browse_main_loop();
	} else
#endif
	{
		if (write_output(outputname))
			return 1;
	}
	return 0;
}

/* Copy file suffix list from optarg to
 * the 2-d array suffilter. */

void parse_filters(char *optarg)
{
	int cf;
	int i;
	int j;
	char tempbuf[NAME_MAX];

	filters = 1;
	for (i=0; i<strlen(optarg); i++) {
		if (optarg[i] == ',')
			filters++;
	}
	suffilter = xmalloc(filters*sizeof (char *));
	for (i=0, cf=0; cf<filters; cf++,i++) {
		j = 0;
		while (optarg[i] != ',' && optarg[i] != '\0' && j < NAME_MAX-1) {
			if (ignorecase)
				tempbuf[j++] = toupper(optarg[i++]);
			else
				tempbuf[j++] = optarg[i++];
		}
		tempbuf[j] = '\0';
		suffilter[cf] = xmalloc(j);
		strcpy(suffilter[cf], tempbuf);
	}
}

/* Returs 1 if string ends in suffix, 0 otherwise. */

int has_suffix(int caseinsensitive, char *string, char *suffix)
{
	int i;

	for (i=0; i<strlen(suffix); i++) {
		if (caseinsensitive) {
			if (toupper(string[strlen(string)-i-1]) != toupper(suffix[strlen(suffix)-i-1]))
				return 0;
		} else {
			if (string[strlen(string)-i-1] != suffix[strlen(suffix)-i-1])
				return 0;
		}
	}
	return 1;
}

/* Check requested access rights for filename.
 * Returns 0 if file can be accessed as requested, 1
 * if; a) permission denied, b) it already exists,
 * we're trying to write and force_overwrite=0,
 * c) it doesn't exist and we're trying to read
 * d) it's a dir and we're trying to write
 * Only prints errors if acc_quiet == 0. */

int check_access(int acc_quiet, int force_overwrite, int mode, char *filename)
{
	struct stat filestats;

	if (!strcmp(filename, "-"))
		return 0;
	if (access(filename, F_OK)) {
		if (mode & R_OK && !(mode & W_OK)) {
			if (!acc_quiet)
				fprintf(stderr, "'%s': file doesn't exist.\n", filename);
			return 1;
		}
		return 0;
	}
	if (stat(filename, &filestats)) {
		if (!acc_quiet)
			perror(filename);
		return 1;
	}
	if (S_ISDIR(filestats.st_mode) && mode & W_OK) {
		if (!acc_quiet)
			fprintf(stderr, "'%s' already exists and is a directory!\n", filename);
		return 1;
	}
	if (mode & W_OK && !force_overwrite) {
		if (!acc_quiet)
			fprintf(stderr, "'%s': file already exists, force overwrite with -F.\n", filename);
		return 1;
	}
	if (access(filename, mode)) {
		if (!acc_quiet)
			perror(filename);
		return 1;
	}
	return 0;
}

/* Screwing up with program invocation gets you here. */

void errmsg(int errormode, char *name)
{
	switch (errormode) {
		case FCULT_INV_TYPE:
			fprintf(stderr, "Invalid list type. Has to be one of:\n0, sld;  1, slx;  2, ddd\n\n");
			break;

		case FCULT_EXTRA_MODE:
			fprintf(stderr, "Only one of the mode commands can be used at a time.\n\n");
			break;

#ifdef HAVE_LIBNCURSES
		case FCULT_BROWSE_STDIN:
			fprintf(stderr, "Reading from stdin currently not possible for the browser -- sorry.\n\n");
			break;

		case FCULT_STDIN_CONF:
			fprintf(stderr, "Can't use stdin as configuration file.\n\n");
			break;
#endif
	}
	fprintf(stderr, "Usage: '%s <-N | -A | -X | -O> [OPTIONS] [-B | -o OUT_FILE] [IN_FILE_1] [...]'\n", name);

	if (errormode > -1)
#ifdef HAVE_GETOPT_H
		fprintf(stderr, "For help, use '%s -h' or '%s --help'\n", name, name);
#else
		fprintf(stderr, "For help, use '%s -h'\n", name);
#endif
	return;
}

/* -h or --help gets you here. */

void show_help(char *name)
{
	fprintf(stderr, "F-CULT does various operations with lists of files.\n\n");

#ifndef HAVE_GETOPT_H
	fprintf(stderr, "Long options are not available in your compile of F-CULT.\n");
#endif

	errmsg(-1, name);
	fprintf(stderr, "\n\
Operation modes (one must be used):\n\
  -N, --not, --missing     output entries not in any of the input files\n\
  -A, --and, --duplicates  output entries in 2 or more of the input files\n\
  -X, --xor, --compare     output entries unique to list # 1\n\
  -O, --or, --combine      output entries from all input files\n\
\n\
Common options:\n\
\n\
  -a, --add-prefix=PREFIX  output 'PREFIX' in front of every dir name\n");
#ifdef HAVE_LIBNCURSES
fprintf(stderr, "\
  -B  --browse             browse the list instead of outputting\n");
#endif
fprintf(stderr, "\
  -b, --bzip2              filter output through bzip2\n\
  -d, --dos-dirs           output '\\' instead of '/' between dir names\n\
  -e, --no-empty-dirs      don't output empty directories\n\
  -F, --force-overwrite    write to file even if it exists\n\
  -f, --filter=SUFFIX,LIST only output file entries with a listed suffix\n\
  -h, --help               show this summary of options and exit\n\
  -I, --ignore-case        ignore case of file entries and filter suffixes\n\
  -i, --ignore-symlinks    ignore symbolic links\n\
  -l, --license            show software license and exit\n\
  -n, --no-subdirs         don't read subdirectories\n\
  -o, --output-to=FILE     write output to FILE\n\
  -q, --quiet              suppress warnings and progress information\n\
  -r, --all-in-root        'move' contents of subdirs to root of the list\n\
  -S, --strip-dirs         strip dir names preceding the listed dir\n\
  -T, --type=TYPE          output type; 0=sld, 1=slx, 2=ddd (default)\n\
  -V, --version            show version information and exit\n\
  -v, --verbose            output each processed dir on screen\n\
  -Z, --zip                filter output through zip\n\
  -z, --gzip               filter output through gzip\n\
\n\
Options specific to the NOT mode:\n\
  -s, --start-from=NUMBER  output files starting from NUMBER\n\
  -t, --allow-trailing     allow one trailing letter in number string\n\
\n");

#ifdef HAVE_LIBNCURSES
	fprintf(stderr, "\
Options for the browser:\n\
\n\
  -c, --config-file=FILE   read site list from FILE (default ~/.fcultrc\n\
  -p, --plain-gfx          use this if screen 'graphics' look messed up\n\
\n");
#endif

#ifdef __MSDOS__
	fprintf(stderr, "For more information see the 'fcultman.dos' file.\n");
#else
	fprintf(stderr, "For more information see the man page (man 1 fcult).\n");
#endif
	return;
}

#define OUT_OF_VMEM fprintf(stderr, "Virtual memory exhausted!\n")

/* calloc() wrapper. Prints an error message and exits
 * if can't allocate memory. If browsing, quits
 * the browser before doing this. Otherwise returns
 * the address calloc() returns. */

void *xcalloc(size_t nmemb, size_t size)
{
	void *addr;

	if (!(addr = calloc(nmemb, size))) {
#ifdef HAVE_LIBNCURSES
		if (browse) {
			endwin();
			putc('\n', stderr);
		}
#endif
		OUT_OF_VMEM;
		exit(1);
	}
	return addr;
}

/* See previous. */

void *xmalloc(size_t size)
{
	void *addr;

	if (!(addr = malloc(size))) {
#ifdef HAVE_LIBNCURSES
		if (browse) {
			endwin();
			putc('\n', stderr);
		}
#endif
		OUT_OF_VMEM;
		exit(1);
	}
	return addr;
}

/* See previous. */

void *xrealloc(void *ptr, size_t size)
{
	void *addr;

	if (!(addr = realloc(ptr, size))) {
#ifdef HAVE_LIBNCURSES
		if (browse) {
			endwin();
			putc('\n', stderr);
		}
#endif
		OUT_OF_VMEM;
		exit(1);
	}
	return addr;
}
