//
// Copyright (C) 1998, Mark W J Redding <grawlfang@ukhackers.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "Alarm.h"

#include "Klook.h"
#include "AlarmDialog.h"
#include "Appointment.h"
#include "docking.h"
#include "KalIO.h"

//
// default constructor
//
Alarm::Alarm(Appointment *parent, Klook* p, const char *n)
	: QObject((QObject*)p,n)
{
	alarmTime = new QDateTime();
	appointment = parent;
	mainW = p;
	dueOption = 0;
	hasFired = false;
	repeating = false;
	alarmBox = 0;
	identifier = 0;
}

//
// constructor for a given date and hour and minute
//
Alarm::Alarm(Appointment *parent, QDate& d, int h, int m, int due, Klook *p , const char *n)
	: QObject((QObject*)p,n)
{
	alarmTime = new QDateTime();
	appointment = parent;
	mainW = p;
	setDate(d);
	setTime(h,m,due);
	dueOption = due;
	hasFired = false;
	repeating = false;
	alarmBox = 0;
	identifier = 0;
}

//
// remove the alarm
//
Alarm::~Alarm()
{
	delete alarmTime;
	if(alarmBox != (AlarmDialog*)0)
	{
		disconnect(alarmBox);
		delete alarmBox;
	}
}

//
// set the date portion of the alarm
//
void
Alarm::setDate(QDate& d)
{
	alarmTime->setDate(d);
}

//
// Start the timer running
//
void
Alarm::goTimer()
{
int secsToGo;
QDateTime now = QDateTime::currentDateTime();

	secsToGo = now.secsTo(*alarmTime);
	if(secsToGo > 0)
	{
		beginTimer(secsToGo);
	}
	else if(!hasFired)
	{
		timerMessage();
	}
}

//
// actually start the timer
//
void
Alarm::beginTimer(int s)
{
	QTimer::singleShot(s * 1000, this, SLOT(timerMessage()));
}

//
// set the time portion of the alarm
//
void
Alarm::setTime(int h, int m, int due)
{
int hour,min,day,month,year,dmon;
int diffHour,diffMin,diffDay;

	diffHour = diffMin = diffDay = 0;

	day = alarmTime->date().day();
	month = alarmTime->date().month();
	year = alarmTime->date().year();
	dmon = alarmTime->date().daysInMonth();

	hour = h;
	min = m;
	
	switch(due)
	{
	case 0: diffMin = 15;	break;
	case 1: diffMin = 30;	break;
	case 2: diffHour = 1;	break;
	case 3: diffHour = 2;	break;
	case 4: diffHour = 3;	break;
	case 5: diffDay = 1;	break;
	case 6: diffDay = 2;	break;
	case 7: diffDay = 3;	break;
	case 8: diffDay = 7;	break;
	}
	
	min -= diffMin;	
	if(min < 0)
	{
		min += 60;
		hour--;
	}

	hour -= diffHour;
	if(hour < 0)
	{
		hour += 24;
		day--;
	}

	day -= diffDay;
	if(day < 0)
	{
		month--;
		if(month == 0)
		{
			month = 12;
			year--;
		}
		QDate tdat(year,month,1);
		dmon = tdat.daysInMonth();
		day += dmon;
	}

	QTime t(hour,min,0);
	alarmTime->setTime(t);
	QDate d(year,month,day);
	alarmTime->setDate(d);
	dueOption = due;
}

//
// return the due option
//
int
Alarm::getDue()
{
	return dueOption;
}

//
// save the alarm
//
void
Alarm::save(bool hasChanged)
{
	if(!appointment->getStream()->isDatabase())
	{
		if(hasChanged)
		{
			appointment->getStream()->write(*alarmTime);
			appointment->getStream()->write(dueOption);
			appointment->getStream()->write(hasFired);
		}
	}
}

//
// load the alarm
//
void
Alarm::load(int version)
{
	if(appointment->getStream()->isDatabase())
	{
	int hour,min;
	QTime tim;

		appointment->getStream()->fetch(5);
		identifier = appointment->getStream()->getNumber(5,0);
		alarmTime->setDate(appointment->getStream()->getDate(5,1));
		hour = appointment->getStream()->getNumber(5,2);
		min = appointment->getStream()->getNumber(5,3);
		tim.setHMS(hour,min,0,0);
		alarmTime->setTime(tim);
		dueOption = appointment->getStream()->getNumber(5,4);
		hasFired = appointment->getStream()->getNumber(5,5) ? true : false;	
	}
	else
	{
		appointment->getStream()->read(alarmTime);
		appointment->getStream()->read(&dueOption);
		appointment->getStream()->read(&hasFired);
	}
}

//
// display a message when the timer expires
//
void
Alarm::timerMessage()
{
	if(alarmBox == (AlarmDialog*)0)
	{
		hasFired = true;
		alarmBox = new AlarmDialog(appointment,mainW);
		alarmBox->setGeometry(0,60,200,200);
		alarmBox->blink();
		connect(alarmBox,SIGNAL(fired()),this,SLOT(killAlarmBox()));
		connect(alarmBox,SIGNAL(repeat()),this,SLOT(reIssueAlarmBox()));
		mainW->markAsChanged();
	}
	else
	{
		alarmBox->blink();
	}
	mainW->getDock().setAlarm(this);
}

//
// destroy the alarm box creates by timerMessage
//
void
Alarm::killAlarmBox()
{
	mainW->getDock().resetAlarm(this);
	alarmBox->hide();
	repeating = false;
}

//
// re-issue an alarm
//
void
Alarm::reIssueAlarmBox()
{
	killAlarmBox();
	repeating = true;
	beginTimer();
}

//
//
//
QString&
Alarm::getSubject()
{
	return appointment->getSubject();
}

//
// Return a reference to the alarm box
//
AlarmDialog*
Alarm::getAlarmBox()
{
	return alarmBox;
}

//
// return the status of the alarm
//
bool
Alarm::alreadyFired()
{
	return hasFired;
}

//
// is the alarm currently in a repeat loop
//
bool
Alarm::isRepeating()
{
	return repeating;
}

int
Alarm::getID()
{
	return identifier;
}

void
Alarm::update()
{
	if(appointment->getStream()->isDatabase())
	{
	QString query;
	QString sDate;

		sDate.sprintf("%04d-%02d-%02d",alarmTime->date().year(),alarmTime->date().month(),alarmTime->date().day());

		if(identifier)
		{
			query.sprintf("update alarm set alarm_date='%s',alarm_hour=%d,alarm_min=%d,dueoption=%d,hasfired=%d where alarm_id=%d",
				sDate.data(),
				alarmTime->time().hour(),
				alarmTime->time().minute(),
				dueOption,
				(hasFired) ? 1 : 0,
				identifier);
		}
		else
		{
			query.sprintf("insert into alarm (alarm_date,alarm_hour,alarm_min,dueoption.hasfired) values ('%s',%d,%d,%d,%d)",
				sDate.data(),
				alarmTime->time().hour(),
				alarmTime->time().minute(),
				dueOption,
				(hasFired) ? 1 : 0);
				
		}
		int id = appointment->getStream()->execute(query);
		if(!identifier)
		{
			identifier = id;
		}
	}
}

void
Alarm::remove()
{
	if(appointment->getStream()->isDatabase())
	{
	QString query;

		query.sprintf("delete from alarm where alarm_id=%d",identifier);
		appointment->getStream()->execute(query);
	}
}

