//
// Copyright (C) 1998, Mark W J Redding <grawlfang@ukhackers.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "Appointment.h"

#include "Alarm.h"
#include "Repeat.h"
#include "Day.h"
#include "Calendar.h"
#include "Export.h"
#include "TimeSlots.h"
#include "KalIO.h"
#include "Klook.h"

//
// create a new appointment with just a start time
// typically, will be called when creating a new appointment
//
Appointment::Appointment(Klook* w, Day* d, int stim)
{
	subject = new QString("");
	text = new QString("");
	location = new QString("");
	if(stim == -1)
	{
		stim = 0;
	}
	setStartTime(stim);
	setFinishTime(stim);
	Sensitivity = 0;
	valid = false;
	alarmed = false;
	mainW = w;
	day = d;
	repeat = (Repeat*)0;
	alarm = (Alarm*)0;
	identifier = 0;
}

//
// create a new appointment
//
Appointment::Appointment(Klook* w, Day* d, QString& sub, int stim , int ftim , QString& txt=0 , QString& loc=0)
{
	subject = new QString(sub);
	text = new QString(txt);
	location = new QString(loc);
	setStartTime(stim);
	setFinishTime(ftim);
	setText(txt);
	Sensitivity = 0;
	valid = false;
	alarmed = false;
	mainW = w;
	day = d;
	repeat = (Repeat*)0;
	alarm = (Alarm*)0;
	identifier = 0;
}

//
// tidy up
//
Appointment::~Appointment()
{
	delete subject;
	delete text;
	delete location;
	if(alarmed)
	{
		clearAlarm();
	}
	if(repeat != (Repeat*)0)
	{
		day->getCalendar()->removeRepeat(repeat);
		delete repeat;
	}
}

//
// set the subject text
//
void
Appointment::setSubject( QString& sub )
{
	subject->resize(0);
	subject->append(sub);
}

//
// return a reference to the subject text
//
QString&
Appointment::getSubject()
{
	return *subject;
}

//
// set the descriptive text
//
void
Appointment::setText( QString& app )
{
	text->resize(0);
	text->append(app);
}

//
// return a reference to the descriptive text
//
QString&
Appointment::getText()
{
	return *text;
}

//
// set the subject text
//
void
Appointment::setLocation( QString& loc )
{
	location->resize(0);
	location->append(loc);
}

//
// return a reference to the location text
//
QString&
Appointment::getLocation()
{
	return *location;
}

//
// set the start time (a value between 0 and TimeSlots::totalSlots)
//
void
Appointment::setStartTime( int t )
{
	startTime = t;
}

//
// set the finish time (a value between 0 and TimeSlots::totalSlots)
//
void
Appointment::setFinishTime( int t )
{
	finishTime = t;
}

//
// return the start time
//
int
Appointment::getStartTime()
{
	return startTime;
}

//
// return the finish time
//
int
Appointment::getFinishTime()
{
	return finishTime;
}

//
// save this appointment to a file
//
void
Appointment::save(bool hasChanged)
{
int len;

	if(!day->getStream()->isDatabase())
	{
		if(hasChanged)
		{
			day->getStream()->write(startTime);
			day->getStream()->write(finishTime);

			len = subject->length();
			day->getStream()->write(len);
			if(len)
			{
				day->getStream()->write(*subject,len);
			}

			len = text->length();
			day->getStream()->write(len);
			if(len)
			{
				day->getStream()->write(*text,len);
			}

			len = location->length();
			day->getStream()->write(len);
			if(len)
			{
				day->getStream()->write(*location,len);
			}

			day->getStream()->write(alarmed);
			day->getStream()->write(Sensitivity);

			if(identifier == 0)
			{
				identifier = mainW->getAppointmentID();
			}
			day->getStream()->write(identifier);

		}

		if(alarmed)
		{
			alarm->save(hasChanged);
		}

		if(hasChanged)
		{
			if(repeat != (Repeat*)0)
			{
				day->getStream()->write(1);
				repeat->save(hasChanged);
			}
			else
			{
				day->getStream()->write(0);
			}
		}
	}
}

//
// load an appointment from a file
//
void
Appointment::load(int version, bool isme)
{
int len;
int repeat_id = 0;
int alarm_id = 0;

	if(day->getStream()->isDatabase())
	{
		day->getStream()->fetch(3);
		identifier = day->getStream()->getNumber(3,0);
		startTime = day->getStream()->getNumber(3,1);
		finishTime = day->getStream()->getNumber(3,2);
		subject->resize(0);
		subject->append(day->getStream()->getString(3,3));
		text->resize(0);
		text->append(day->getStream()->getString(3,4));
		location->resize(0);
		location->append(day->getStream()->getString(3,5));
		Sensitivity = day->getStream()->getNumber(3,6);
		repeat_id = day->getStream()->getNumber(3,7);
		alarm_id = day->getStream()->getNumber(3,8);	
	}
	else
	{
		day->getStream()->read(&startTime);
		day->getStream()->read(&finishTime);
		if(version < 13)
		{
			startTime *= 2;
			finishTime *= 2;
		}

		day->getStream()->read(&len);
		subject->resize(len);
		if(len)
		{
			day->getStream()->read(subject,len);
		}
		subject->truncate(len);

		day->getStream()->read(&len);
		text->resize(len);
		if(len)
		{
			day->getStream()->read(text,len);
		}
		text->truncate(len);

		if(version > 13)
		{
			day->getStream()->read(&len);
			location->resize(len);
			if(len)
			{
				day->getStream()->read(location,len);
			}
			location->truncate(len);
		}

		day->getStream()->read(&alarmed);
		if(version > 1)
		{
			day->getStream()->read(&Sensitivity);
		}
		else
		{
			Sensitivity = public_access;
		}

		if(version > 13)
		{
			day->getStream()->read(&identifier);
		}
	}

	makeValid(mainW);

	if(alarmed)
	{
		alarm = new Alarm(this,mainW);
		if(day->getStream()->isDatabase())
		{
		QString query;

			query.sprintf("select alarm_id,alarm_date,alarm_hour,alarm_min,dueoption,hasfired from alarm where alarm_id=%d",alarm_id);
			day->getStream()->select(5,query);
		}

		alarm->load(version);

		if(isme)
		{
			alarm->goTimer();
		}
		if(day->getStream()->isDatabase())
		{
			day->getStream()->doneSelect(5);
		}
	}

	if(day->getStream()->isDatabase())
	{
		if(repeat_id)
		{
                QString query;

                        query.sprintf("select repeat_id,enddate,waittime from repeat where repeat_id=%d",repeat_id);
                        day->getStream()->select(6,query);

                        repeat = new Repeat(day,this);
                        repeat->load(version);
                        day->getCalendar()->addRepeat(repeat);
			day->getStream()->doneSelect(6);
		}
		else
		{
			repeat = (Repeat*)0;
		}
	}
	else
	{
		if(version > 7)
		{
		int rep;

			day->getStream()->read(&rep);
			if(rep == 1)
			{
				repeat = new Repeat(day,this);
				repeat->load(version);
				day->getCalendar()->addRepeat(repeat);
			}
		}
		else
		{
			repeat = (Repeat*)0;
		}
	}
}

//
// return the status of an appointment
// in order to use the EditAppointment class an instance of an appointment 
// must exist. It is created as 'new' (or not valid). When the user exits
// from the EditAppointment class the instance of the appointment can be
// marked as value (see next method). If the user quits from EditAppointment 
// then the instance of the appointment should be deleted
//
bool
Appointment::isNew()
{
	return !valid;
}

//
// make an appointment valid
//
void
Appointment::makeValid(Klook* w)
{
	valid = true;
	mainW = w;
}

//
// Set an alarm
//
void
Appointment::setAlarm(int due, QDate& day)
{
int h,m;

	h = startTime / TimeSlots::hourUnits;
	m = (startTime % TimeSlots::hourUnits) * TimeSlots::minuteUnits;

	if(!alarmed)
	{
		alarm = new Alarm(this,day,h,m,due,mainW);
	}
	alarmed = true;
	alarm->goTimer();
}

//
// Clear the alarm
//
void
Appointment::clearAlarm()
{
	delete alarm;
	alarmed = false;
	alarm = (Alarm*)0;
}

//
// return the status of any alarm
// 
bool
Appointment::isAlarmed()
{
	return alarmed;
}

//
// return a reference to the alarm
//
Alarm&
Appointment::getAlarm()
{
	return *alarm;
}

//
// return the status of an alarm (if it exists)
//
int
Appointment::getAlarmStatus()
{
int rs = none;

	if(alarmed)
	{
		rs = armed;
		if(alarm->alreadyFired())
		{
			rs = fired;
		}
		if(alarm->isRepeating())
		{
			rs = repeating;
		}
	}
	return rs;
}

//
// set the sensitivity level
//
void
Appointment::setSensitivity(int s)
{
	Sensitivity = s;
}

//
// return the sensitivity level
//
int
Appointment::getSensitivity()
{
	return Sensitivity;
}

void
Appointment::setRepeat(Repeat* r)
{
	repeat = r;
}

Repeat*
Appointment::getRepeat()
{
	return repeat;
}

bool
Appointment::isRepeating()
{
	return (repeat != (Repeat*)0) ? true : false;
}

void
Appointment::clearRepeat()
{
	if(repeat != (Repeat*)0)
	{
		day->getCalendar()->removeRepeat(repeat);
		repeat->remove();
		delete repeat;
	}
	repeat = (Repeat*)0;
}

Day*
Appointment::getDay()
{
	return day;
}


//
// search the subject and text for a string
//
bool
Appointment::findString(QString& str)
{
	return ((subject->find(str.data(),0,false) != -1) || (text->find(str.data(),0,false) != -1)) ? true : false;
}


//
// export the appointment
//
void
Appointment::export(QTextStream& fil, Export& expdata)
{
int len;
char *add;
QString d;
TimeSlots ts;

	fil << expdata.delimiter();
	fil << "A";	// mark it as an appointment
	fil << expdata.delimiter();

	fil << ts.whichTime(startTime);
	fil << expdata.delimiter();

	fil << ts.whichTime(finishTime);
	fil << expdata.delimiter();

	len = subject->length();
	add = subject->data();
	if(len)
	{
		fil.writeRawBytes(add,len);
	}
	else
	{
		fil << " ";
	}
	fil << expdata.delimiter();

	len = text->length();
	add = text->data();
	if(len)
	{
		fil.writeRawBytes(add,len);
	}
	else
	{
		fil << " ";
	}
	fil << expdata.delimiter();

	len = location->length();
	add = location->data();
	if(len)
	{
		fil.writeRawBytes(add,len);
	}
	else
	{
		fil << " ";
	}
	fil << expdata.delimiter();

	d.sprintf("%d",Sensitivity);
	fil << d;
//	fil << expdata.delimiter();

//	d.sprintf("%d",(int)alarmed);
//	fil << d;
}

int
Appointment::getID()
{
	return identifier;
}

void
Appointment::setID(int id)
{
	identifier = id;
}

void
Appointment::update()
{
	if(day->getStream()->isDatabase())
	{
	QString query;
printf("id was %d\n",identifier);
		if(identifier)
		{
			query.sprintf("update appointment set starttime=%d,finishtime=%d,subject='%s',text='%s',location='%s',sensitivity=%d,repeat_id=%d,alarm_id=%d where appointment_id=%d",
				startTime,
				finishTime,
				subject->data(),
				text->data(),
				location->data(),
				Sensitivity,
				(repeat) ? repeat->getID() : 0,
				(alarm) ? alarm->getID() : 0,
				identifier);
		}
		else
		{
			query.sprintf("insert into appointment (user_id,day_id,starttime,finishtime,subject,text,location,sensitivity,repeat_id,alarm_id) values (%d,%d,%d,%d,'%s','%s','%s',%d,%d,%d)",
				day->getStream()->getUserID(),
				day->getID(),
				startTime,
				finishTime,
				subject->data(),
				text->data(),
				location->data(),
				Sensitivity,
				(repeat != (Repeat*)0) ? repeat->getID() : 0,
//				(alarm != (Alarm*)0) ? alarm->getID() : 0);
				(alarmed) ? alarm->getID() : 0);
		}
		int id = day->getStream()->execute(query);
		if(!identifier)
		{
			identifier = id;
		}
	}
}

KalIO*
Appointment::getStream()
{
	return day->getStream();
}

void
Appointment::remove()
{
	if(getStream()->isDatabase())
	{
	QString query;

		query.sprintf("delete from appointment where appointment_id=%d",identifier);
		getStream()->execute(query);

		if(repeat != (Repeat*)0)
		{
			repeat->remove();
		}

		if(alarmed)
		{
			query.sprintf("delete from alarm where alarm_id=%d",alarm->getID());
			getStream()->execute(query);
		}
	}
}

