/* Copyright (C) 1999 Donald J. Bindner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/* I would like to acknowledge Jouni.Lohikoski@iki.fi whose utility
 * tleds influenced some of the look and feel for this program and
 * whose source code served as my example for process-handling.
 */

#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
/* #include <fcntl.h> */

#define VERSION "0.3"
#define MYNAME "mailblink"
#define _PATH_TMP "/tmp"

/* Various statuses we return after checking for a PID file */
#define PID_INVALID -1
#define PID_NONE 0
#define PID_WE_OWN 1
#define PID_OTHER_OWNER 2
#define PID_FILE_MISSING 3

static char pidDir[40];		/* 40 should be enough */
static char pidFileName[80];	/* 80 should be enough */
char *ledProgram = NULL;
static pid_t child;		/* for forking */
static int is_blinking = 0;	/* flag so we can keep track */
static char *light[3] = { "num", "caps", "scroll" };
static char *ledPath[] = { 
    "/usr/local/lib/mailblink", 
    "/usr/lib/mailblink", 
    NULL };
static int opt_k=0, opt_q=0, opt_D;	/* some user options */

void help_exit( void ) {
    printf( "%s version %s\n\
Usage: %s [options]\n\
\t-d <delay>\tDelay between checks (range: 10 to 3600 seconds)\n\
\t-f <file>\tMail file to watch.\n\
\t-k\t\tKill old processes.\n\
\t-l <led>\tLed to blink (num, caps or scroll)\n\
\t-q\t\tQuiet.  Don't print any messages.\n\
\t-D\t\tDebug.  Don't fork into the background.\n", MYNAME, VERSION, MYNAME );
    exit( 1 );
}

void message( const char *format, ... ) {
    va_list ap;

    if( opt_q ) return;

    va_start( ap, format );
    vprintf( format, ap );
}

void warning( const char *format, ... ) {
    va_list ap;

    if( opt_q ) return;

    va_start( ap, format );
    vfprintf( stderr, format, ap );
}

int pid_file_status( void ) {
/*
 * Checks for a PID file and returns a status based on what it finds.
 */
    struct stat status;

    if( lstat( pidDir, &status )) { 
	/* directory isn't there so the pidfile can't be either */
	return PID_NONE;
    } else {
	/* invalid if it is not a directory */
	if( !S_ISDIR( status.st_mode )) return PID_INVALID;
	/* invalid if anyone else but owner can write to it */
	if( status.st_mode & ( S_IWGRP | S_IWOTH )) return PID_INVALID;
    }

    if( stat( pidFileName, &status )) {
	return PID_FILE_MISSING;
    }

    if( status.st_uid == getuid() ) return PID_WE_OWN;

    return PID_OTHER_OWNER;
}

void cleanup_dead_pid( pid_t pid, int light_num ) {
/* 
 * Called to remove the PID file.  First checks to make
 * sure that the process no longer exists.
 */
    char procFileName[80];
    pid_t oldPid;
    struct stat status;

    oldPid = get_old_pid();
    
    if( !oldPid ) {
	warning( "Can't find old process to clean up.\n" );
	warning( "%s may need to be removed by hand.\n", pidDir );
	kill( pid, SIGTERM );
	exit( 1 );
    }
    
    sprintf(procFileName, "/proc/%ld/environ", (long)oldPid);
    if( !stat(procFileName, &status )) {	/* The old proc. */
	warning( "%s: %s lock seems to be taken already by PID=%ld\n",
		MYNAME, light[light_num], (long)oldPid );
	kill( pid, SIGTERM );
	exit( 1 );
    } else {
	/* The old process was not alive, so we try to fix. */
	unlink( pidFileName );
	rmdir( pidDir );

	if( pid_file_status() != PID_NONE ) {
	    warning( "%s: Can't remove %s\n", MYNAME, pidDir );
	    warning( "Program terminated.\n");
	    kill( pid, SIGTERM );
	    exit( 1 );
	}
    }
}

void create_pid_file( pid_t pid, int light_num ) {
/*
 * Creates our PID file.  Checks of course to be sure one
 * doesn't exist from before.  But if one does, it will
 * remove it if the process no longer exists.
 */
    mode_t old_umask;
    FILE *f;

    /* Check to see if it already exists */
    switch( pid_file_status()) {
	case PID_INVALID:
	    warning( "Something is wrong with %s or %s.\n", 
		    pidDir, pidFileName );
	    warning( "Check for existing processes and\n" );
	    warning( "remove %s before trying again.\n", pidDir );
	    kill( pid, SIGTERM );
	    exit( 1 );
	    break;
	case PID_FILE_MISSING:
	    warning( "Cannot create PID file %s\n", pidFileName );
	    warning( "Remove %s before trying again.\n", pidDir );
	    kill( pid, SIGTERM );
	    exit( 1 );
	    break;
	case PID_OTHER_OWNER:
	    warning( "Cannot create PID file %s\n", pidFileName );
	    warning( "One already exists from another user.\n" );
	    kill( pid, SIGTERM );
	    exit( 1 );
	    break;
	case PID_WE_OWN:
	    cleanup_dead_pid( pid, light_num );
	    /* If this returns we can fall through to the next part. */
	case PID_NONE:
	    break; /* Break out of the switch and do it */
	default:
	    abort();
    }

    /* Set umask to something safe and friendly */
    old_umask = umask( 022 );
    
    /* Open a directory to hold our PID file */
    if( mkdir( pidDir, 0755 )) {
	warning( "Unable to create directory %s.  Dying.\n", pidDir );
	kill( pid, SIGTERM );
	exit( 1 );
    }

    /* Open the PID file finally */
    if(( f = fopen( pidFileName, "w" )) == NULL ) {
	rmdir( pidDir );
	warning( "Unable to create file %s.  Dying.\n", pidFileName );
	kill( pid, SIGTERM );
	exit( 1 );
    }

    /* Write our PID in the file */
    fprintf( f, "%ld\n", (long)pid );

    /* Close the file and check for any last problems */
    if( fclose( f )) {
	unlink( pidFileName );
	rmdir( pidDir );
	warning( "Unable to close %s.  Dying.\n", pidFileName );
	kill( pid, SIGTERM );
	exit( 1 );
    }
}

pid_t get_old_pid() {
/*
 * Gets the old PID by reading it from the PID file.
 * It is assumed that a valid PID file already exists.
 */
    FILE *pidFile;
    long returnValue = 0;

    if( !( pidFile = fopen( pidFileName, "r" ))) {
	return (pid_t)0L;
    }
    fscanf(pidFile, "%ld", &returnValue);
    fclose(pidFile);
    return (pid_t) returnValue;
}

int kill_old_process() {
/*
 * Will kill the (old) process who's PID is read from a
 * valid PID file.
 */
    pid_t pid, pid2;

    switch( pid_file_status() ) {
	case PID_INVALID:
	    warning( "Something is wrong with %s or %s.\n", 
		    pidDir, pidFileName );
	    warning( "Won't kill anything.\n" );
	    return 1;
	    break;
	case PID_FILE_MISSING:
	    warning( "%s file missing or invalid.\n", pidFileName );
	    warning( "Couldn't find what to kill.\n" );
	    return 1;
	    break;
	case PID_OTHER_OWNER:
	    warning( "%s file owned by another user.\n", pidFileName );
	    warning( "Won't kill process even if it exists.\n" );
	    return 1;
	    break;
	case PID_NONE:
	    warning( "Couldn't find what to kill.\n" );
	    return 1;
	    break;
	case PID_WE_OWN:
	    if( !( pid = get_old_pid())) {
		warning( "Couldn't find what to kill.\n" );
		return 1;
	    }
	    /* ready to kill -- drop out of the switch here */
	    break;
	default:
	    abort();
    }

    kill( pid, SIGTERM );
    message( "One moment...(3 secs)...\n" );
    sleep( 3 );

    if( pid2 = get_old_pid()) {
	warning( "PID: %ld - Hmm...not sure if I succeeded in kill.\n", 
		(long)pid2 );
	return 1;
    }

    message( "Killed PID %ld.\n", (long)pid );

    return 0;
}

void kill_blinking( void ) {
/*
 * This function kills the "led" process when we want the blinking
 * to stop.
 */
    if( !is_blinking ) return;

    kill( child, SIGINT );
    while( child != waitpid( child, NULL, 0 ));

    is_blinking = 0;
}

void set_blinking( int led, int num_blinks ) {
/* 
 * This function forks off and loads the external program
 * "led" that actually blinks the led for us.
 */
    char st_led[] = "...";
    char st_blinks[] = ".......";
    struct stat status;

    sprintf( st_led, "%d", led );
    sprintf( st_blinks, "%d", num_blinks );

    if( is_blinking ) kill_blinking();

    if( stat( ledProgram, &status )) {
	warning( "%s: cannot find %s.\n", MYNAME, ledProgram );
	exit( 1 );
    }

    if( child = fork()) { /* parent code */
	if( child != -1 ) is_blinking = 1;

	return;
    }

    /* child code */
    execl( ledProgram, "led", st_led, st_blinks, NULL );  /* call the blinker */
}

void my_exit( void ) {
    kill_blinking();
    if( getpid() == get_old_pid()) {
	unlink( pidFileName );
	rmdir( pidDir );
    }
}

void my_handler( int sig ) {
    exit( 1 );
}

char *find_ledProgram( void ) {
/*
 * Searches the ledPath until it finds a copy of the program
 * called led and returns the full pathname of the program.
 */
    int i;
    char *name, cwd[128];
    struct stat stat_buf;

    /* check each part of the path */
    for( i = 0; ledPath[i] != NULL ; i++ ) {
	if(!( name = malloc( strlen( ledPath[i] ) + 5 ))) {
	    warning( "%s: cannot malloc.\n", MYNAME );
	    exit( 1 );
	}
	strcpy( name, ledPath[i] );
	strcat( name, "/led" );

	if( !stat( name, &stat_buf )) return name;

	free( name );	/* not found -- free the mem and loop */
    }

    /* Couldn't find it in the path check current dir */
    if( NULL == getcwd( cwd, sizeof( cwd ))) return NULL;

    if( !( name = malloc( strlen( cwd ) + 5 ))) {
	warning( "%s: cannot malloc.\n", MYNAME );
	exit( 1 );
    }
    strcpy( name, cwd );
    strcat( name, "/led" );

    if( !stat( name, &stat_buf )) {
	warning( "%s: using led from the current directory.\n", MYNAME );
	return name;
    }

    free( name );
    return NULL;
}

main( int argc, char *argv[] ) {
    int c, i, light_num=2;
    int delay=30, count=0;
    char *filename;
    struct stat stat_buf;
    time_t old_mtime;
    pid_t pid;
    sigset_t sigset;
    
    /* block signals from delivery until we are ready to handle them */
    assert( !sigemptyset( &sigset ));
    assert( !sigaddset( &sigset, SIGTERM ));
    assert( !sigaddset( &sigset, SIGINT ));
    assert( !sigaddset( &sigset, SIGHUP ));
    assert( !sigaddset( &sigset, SIGQUIT ));
    assert( !sigaddset( &sigset, SIGTTOU ));
    assert( !sigprocmask( SIG_BLOCK, &sigset, NULL ));

    /* We'll start with the file from $MAIL but let the user overide */
    filename = getenv( "MAIL" );

    while( EOF != ( c = getopt( argc, argv, "Dd:f:hkl:q" ))) {
	switch( c ) {
	    case 'D':
		opt_D = 1;
		break;
	    case 'd':
		delay = strtol( optarg, NULL, 10 );
		break;
	    case 'f':
		filename = optarg;
		break;
	    case 'h':
		help_exit();
		break;
	    case 'k':
		opt_k = 1;
		break;
	    case 'l':
		light_num = -1;
		for( i = 0; i < 3; i++ ) {
		    if( !strcmp( optarg, light[i] )) light_num = i;
		}
		if( light_num == -1 ) help_exit();
		break;
	    case 'q':
		opt_q = 1;
		break;
	    default:
		help_exit();
	}
    }

    /* Let's make sure that delay has reasonable defaults */
    if( delay > 3600 ) delay = 3600;
    if( delay < 10 ) delay = 10;

    /* Our dir in /tmp */
    sprintf( pidDir, "%s/%s-%s", _PATH_TMP, MYNAME, light[light_num] );
    /* This is our .pid file */
    sprintf( pidFileName, "%s/%s.pid", pidDir, MYNAME );

    if( opt_k ) return kill_old_process();

    /* find an led program to spawn when we need it */
    ledProgram = find_ledProgram();
    if( ledProgram == NULL ) {
	warning( "Can't find led program in %s.\n", ledPath[0] );
	exit( 1 );
    }

    if( filename == NULL ) {
	warning( "Don't know what file to monitor.  Specify a file with -f.\n" );
	exit( 1 );
    } else {
	message( "Preparing to monitor %s on %s lock.\n", filename, light[light_num] );
    }

    if( opt_D ) {	/* Run in foreground to debug */
	message("Debug mode--running in foreground.\n" );
	pid = getpid();
    } else {	/* fork into the background */
	if( (pid = fork()) == -1 ) {
	    warning( "%s: fork() failed", MYNAME );
	    return 1;
	}
    }

    if( pid ) {	/* parent code */
        create_pid_file( pid, light_num );
	message( "%s running with PID %ld.\n", MYNAME, pid );
	if( !opt_D ) exit( 0 );		/* only exit if not debugging */
    }

    /* detatch ourself from a controlling terminal and mounted filesystems */
    if( !opt_D ) {	/* if not debugging */
	setsid();
	chdir( "/" );
	umask( 0 );
    }

    /* Set up our exit protocol */
    assert( !atexit(my_exit));

    /* set signal handler and deliver all blocked signals */
    {
	struct sigaction act;

	memset( &act, 0, sizeof( struct sigaction ));
	act.sa_handler = my_handler;
	act.sa_flags |= SA_RESTART;
	sigfillset( &act.sa_mask );
	assert( !sigaction( SIGTERM, &act, NULL ));
	assert( !sigaction( SIGINT, &act, NULL ));
	assert( !sigaction( SIGHUP, &act, NULL ));
	assert( !sigaction( SIGQUIT, &act, NULL ));
	assert( !sigaction( SIGTTOU, &act, NULL ));
	assert( !sigprocmask( SIG_UNBLOCK, &sigset, NULL ));
    }

    while(1) {
	if( stat( filename, &stat_buf )) {	/* file does not exist */
	    if( count ) {
		count = 0;
		kill_blinking();
	    }
	} else { 				/* file exists */
	    if( stat_buf.st_atime < stat_buf.st_mtime ) {
		if( !count || ( stat_buf.st_mtime > old_mtime )) {
		    count++;
		    old_mtime = stat_buf.st_mtime;
		    set_blinking( light_num+1, count );
		}
	    } else if( count ) {		/* has been read */
		count = 0;
		kill_blinking();
	    }
	}
	sleep( delay );
    }
}

